import json
import asyncio
from mlserver.model import MLModel
from mlserver.settings import ModelSettings
from mlserver.types import (
    InferenceRequest,
    InferenceResponse,
)
from mlserver.codecs import (
    StringCodec,
)
from mlserver_huggingface.common import (
    HuggingFaceSettings,
    parse_parameters_from_env,
    InvalidTranformerInitialisation,
    load_pipeline_from_settings,
    NumpyEncoder,
    SUPPORTED_OPTIMUM_TASKS,
)
from mlserver_huggingface.codecs import MultiStringRequestCodec
from transformers.pipelines import SUPPORTED_TASKS
from mlserver.logging import logger


class HuggingFaceRuntime(MLModel):
    """Runtime class for specific Huggingface models"""

    def __init__(self, settings: ModelSettings):
        env_params = parse_parameters_from_env()
        if not env_params and (
            not settings.parameters or not settings.parameters.extra
        ):
            raise InvalidTranformerInitialisation(
                500,
                "Settings parameters not provided via config file nor env variables",
            )

        extra = env_params or settings.parameters.extra  # type: ignore
        self.hf_settings = HuggingFaceSettings(**extra)  # type: ignore

        if self.hf_settings.task not in SUPPORTED_TASKS:
            raise InvalidTranformerInitialisation(
                500,
                (
                    f"Invalid transformer task: {self.hf_settings.task}."
                    f" Available tasks: {SUPPORTED_TASKS.keys()}"
                ),
            )

        if self.hf_settings.optimum_model:
            if self.hf_settings.task not in SUPPORTED_OPTIMUM_TASKS:
                raise InvalidTranformerInitialisation(
                    500,
                    (
                        f"Invalid transformer task for "
                        f"OPTIMUM model: {self.hf_settings.task}. "
                        f"Supported Optimum tasks: {SUPPORTED_OPTIMUM_TASKS.keys()}"
                    ),
                )

        if settings.max_batch_size != self.hf_settings.batch_size:
            logger.warning(
                f"hf batch_size: {self.hf_settings.batch_size} is different "
                f"from MLServer max_batch_size: {settings.max_batch_size}"
            )

        super().__init__(settings)

    async def load(self) -> bool:
        # Loading & caching pipeline in asyncio loop to avoid blocking
        print("loading model...")
        await asyncio.get_running_loop().run_in_executor(
            None, load_pipeline_from_settings, self.hf_settings
        )
        print("(re)loading model...")
        # Now we load the cached model which should not block asyncio
        self._model = load_pipeline_from_settings(self.hf_settings)
        print("model has been loaded!")
        self.ready = True
        return self.ready

    async def predict(self, payload: InferenceRequest) -> InferenceResponse:
        """
        TODO
        """

        # Adding some logging as hard to debug given the many types of input accepted
        logger.debug("Payload %s", payload)

        # TODO: convert and validate?
        kwargs = self.decode_request(payload, default_codec=MultiStringRequestCodec)

        args = kwargs.pop("args", None)

        X = kwargs.pop("array_inputs", None)
        if X is not None:
            args = [list(X)] + args
        if args is None:
            prediction = self._model(**kwargs)
        else:
            prediction = self._model(args, **kwargs)

        logger.debug("Prediction %s", prediction)

        # TODO: Convert hf output to v2 protocol, for now we use to_json
        if isinstance(prediction, dict):
            str_out = [json.dumps(prediction, cls=NumpyEncoder)]
        else:
            str_out = [json.dumps(pred, cls=NumpyEncoder) for pred in prediction]
        prediction_encoded = StringCodec.encode_output(payload=str_out, name="output")

        return InferenceResponse(
            model_name=self.name,
            model_version=self.version,
            outputs=[prediction_encoded],
        )



import os
import json
from typing import Optional, Dict
from distutils.util import strtobool

import numpy as np
from pydantic import BaseSettings
from mlserver.errors import MLServerError

from transformers.pipelines import pipeline
from transformers.pipelines.base import Pipeline
from transformers.models.auto.tokenization_auto import AutoTokenizer

try:
    # Optimum 1.7 changed the import name from `SUPPORTED_TASKS` to
    # `ORT_SUPPORTED_TASKS`.
    # We'll try to import the more recent one, falling back to the previous
    # import name if not present.
    # https://github.com/huggingface/optimum/blob/987b02e4f6e2a1c9325b364ff764da2e57e89902/optimum/pipelines/__init__.py#L18
    from optimum.pipelines import ORT_SUPPORTED_TASKS as SUPPORTED_OPTIMUM_TASKS
except ImportError:
    from optimum.pipelines import SUPPORTED_TASKS as SUPPORTED_OPTIMUM_TASKS


HUGGINGFACE_TASK_TAG = "task"

ENV_PREFIX_HUGGINGFACE_SETTINGS = "MLSERVER_MODEL_HUGGINGFACE_"
HUGGINGFACE_PARAMETERS_TAG = "huggingface_parameters"
PARAMETERS_ENV_NAME = "PREDICTIVE_UNIT_PARAMETERS"


class InvalidTranformerInitialisation(MLServerError):
    def __init__(self, code: int, reason: str):
        super().__init__(
            f"Huggingface server failed with {code}, {reason}",
            status_code=code,
        )


class HuggingFaceSettings(BaseSettings):
    """
    Parameters that apply only to alibi huggingface models
    """

    class Config:
        env_prefix = ENV_PREFIX_HUGGINGFACE_SETTINGS

    task: str = ""
    pretrained_model: Optional[str] = None
    pretrained_tokenizer: Optional[str] = None
    optimum_model: bool = False
    device: int = -1
    batch_size: Optional[int] = None


def parse_parameters_from_env() -> Dict:
    """
    TODO
    """
    parameters = json.loads(os.environ.get(PARAMETERS_ENV_NAME, "[]"))

    type_dict = {
        "INT": int,
        "FLOAT": float,
        "DOUBLE": float,
        "STRING": str,
        "BOOL": bool,
    }

    parsed_parameters = {}
    for param in parameters:
        name = param.get("name")
        value = param.get("value")
        type_ = param.get("type")
        if type_ == "BOOL":
            parsed_parameters[name] = bool(strtobool(value))
        else:
            try:
                parsed_parameters[name] = type_dict[type_](value)
            except ValueError:
                raise InvalidTranformerInitialisation(
                    "Bad model parameter: "
                    + name
                    + " with value "
                    + value
                    + " can't be parsed as a "
                    + type_,
                    reason="MICROSERVICE_BAD_PARAMETER",
                )
            except KeyError:
                raise InvalidTranformerInitialisation(
                    "Bad model parameter type: "
                    + type_
                    + " valid are INT, FLOAT, DOUBLE, STRING, BOOL",
                    reason="MICROSERVICE_BAD_PARAMETER",
                )
    return parsed_parameters


def load_pipeline_from_settings(hf_settings: HuggingFaceSettings) -> Pipeline:
    """
    TODO
    """
    # TODO: Support URI for locally downloaded artifacts
    # uri = model_parameters.uri
    model = hf_settings.pretrained_model
    tokenizer = hf_settings.pretrained_tokenizer
    device = hf_settings.device

    if model and not tokenizer:
        tokenizer = model

    if hf_settings.optimum_model:
        optimum_class = SUPPORTED_OPTIMUM_TASKS[hf_settings.task]["class"][0]
        model = optimum_class.from_pretrained(
            hf_settings.pretrained_model,
            from_transformers=True,
        )
        tokenizer = AutoTokenizer.from_pretrained(tokenizer)
        # Device needs to be set to -1 due to known issue
        # https://github.com/huggingface/optimum/issues/191
        device = -1

    pp = pipeline(
        hf_settings.task,
        model=model,
        tokenizer=tokenizer,
        device=device,
        batch_size=hf_settings.batch_size,
    )

    # If batch_size > 0 we need to ensure tokens are padded
    if hf_settings.batch_size:
        pp.tokenizer.pad_token_id = [str(pp.model.config.eos_token_id)]  # type: ignore

    return pp


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



from typing import Optional, Type, Any, Dict
from mlserver.codecs.utils import (
    has_decoded,
    _save_decoded,
    get_decoded_or_raw,
)
from mlserver.codecs.base import (
    RequestCodec,
    register_request_codec,
    InputCodec as InputCodecTy,
)
from mlserver.codecs import StringCodec
from mlserver.types import (
    InferenceRequest,
    InferenceResponse,
)


class MultiInputRequestCodec(RequestCodec):
    """
    The SingleInputRequestCodec can be used as a "meta-implementation" for other
    codecs. Its goal to decode the whole request simply as the first decoded
    element.
    """

    InputCodec: Optional[Type[InputCodecTy]] = None

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        if cls.InputCodec is None:
            return False

        return cls.InputCodec.can_encode(payload)

    @classmethod
    def encode_response(
        cls,
        model_name: str,
        payload: Dict[str, Any],
        model_version: Optional[str] = None,
        **kwargs
    ) -> InferenceResponse:
        return InferenceResponse(
            model_name=model_name,
            model_version=model_version,
            outputs=[
                cls.InputCodec.encode_output(key, value, **kwargs)  # type: ignore
                for key, value in payload.items()
            ],
        )

    @classmethod
    def decode_response(cls, response: InferenceResponse) -> Dict[str, Any]:
        values = {}
        for item in response.outputs:
            if not has_decoded(item) and cls.InputCodec is not None:
                decoded_payload = cls.InputCodec.decode_output(item)  # type: ignore
                _save_decoded(item, decoded_payload)

            value = get_decoded_or_raw(item)
            values[item.name] = value
        return values

    @classmethod
    def encode_request(cls, payload: Dict[str, Any], **kwargs) -> InferenceRequest:
        return InferenceRequest(
            inputs=[
                cls.InputCodec.encode_input(key, value, **kwargs)  # type: ignore
                for key, value in payload.items()
            ],
        )

    @classmethod
    def decode_request(cls, request: InferenceRequest) -> Dict[str, Any]:
        values = {}
        for item in request.inputs:
            if not has_decoded(item) and cls.InputCodec is not None:
                decoded_payload = cls.InputCodec.decode_input(item)  # type: ignore
                _save_decoded(item, decoded_payload)

            value = get_decoded_or_raw(item)
            values[item.name] = value
        return values


@register_request_codec
class MultiStringRequestCodec(MultiInputRequestCodec):
    InputCodec = StringCodec
    ContentType = StringCodec.ContentType
